"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PackageSourceV4 = void 0;
const fs_1 = __importDefault(require("fs"));
const url_1 = __importDefault(require("url"));
const path_1 = __importDefault(require("path"));
const del_1 = __importDefault(require("del"));
const cachedir_1 = __importDefault(require("cachedir"));
const builtin_modules_1 = __importDefault(require("builtin-modules"));
const cjs_module_lexer_1 = __importDefault(require("cjs-module-lexer"));
const es_module_lexer_1 = __importDefault(require("es-module-lexer"));
const esbuild = __importStar(require("esbuild"));
const tiny_glob_1 = __importDefault(require("tiny-glob"));
const util_1 = require("../util");
const BROWSER_SUBPATHS = new Set(['browser', 'import', 'default']); // earlier order takes priority (also "default" is required!)
const NODE_ESM_SUBPATHS = new Set(['import', 'default']);
const NODE_CJS_SUBPATHS = new Set(['require', 'node', 'default']);
const NODE_BUILTINS = new Set([...builtin_modules_1.default]);
const ESBUILD_EXTENSIONS = new Set(['.cjs', '.js', '.json', '.jsx', '.mjs', '.ts', '.tsx']);
class PackageSourceV4 {
    constructor(config) {
        this.target = 'browser';
        this.mode = 'development';
        this.pkgPrefix = '/_snowpack/pkg/';
        /** packages to always bundle (i.e. don’t mark as "external") */
        this.alwaysBundle = new Set();
        // load config
        this.config = config;
        // cache
        if (config.buildOptions.cacheDirPath) {
            this.cacheDir = new URL(`file://${path_1.default.resolve(config.buildOptions.cacheDirPath)}/`);
        }
        else {
            this.cacheDir = new URL(`file://${cachedir_1.default('snowpack')}/`);
        }
        // target
        if (config.buildOptions.ssr) {
            this.target = 'node-cjs'; // should "node-esm" ever be used?
        }
        // mode
        if (config.mode === 'production') {
            this.mode = 'production';
            process.env.NODE_ENV = 'production';
        }
        // prefix
        this.pkgPrefix = config.buildOptions.metaUrlPath + '/pkg/';
        // imports
        const importMapLoc = new URL('./import-map.json', this.cacheDir);
        if (fs_1.default.existsSync(importMapLoc)) {
            this.importMap = JSON.parse(fs_1.default.readFileSync(importMapLoc, 'utf8')); // load and cache if exists on disk
        }
        else {
            this.importMap = { imports: {} }; // if new, create new one to be written later
        }
    }
    async prepare() { }
    async prepareSingleFile(fileLoc) {
        console.log({ single: url_1.default.fileURLToPath(fileLoc) });
    }
    async load(id) {
        const spec = id.replace(new RegExp(`^${this.pkgPrefix}`), '');
        const NotFoundError = new Error(`Could not locate "${spec}"`);
        const resolved = this.importMap.imports[spec];
        if (!resolved)
            throw NotFoundError;
        const filename = new URL(resolved, this.cacheDir);
        const contents = await fs_1.default.promises.readFile(filename);
        return {
            contents: await this.transformImports({ filename, contents }),
            imports: [],
        };
    }
    async resolvePackageImport(originalSpec) {
        let spec = originalSpec;
        // 0. if node built-in, return
        if (NODE_BUILTINS.has(originalSpec))
            return originalSpec;
        // 1. handle aliases
        const aliasEntry = util_1.findMatchingAliasEntry(this.config, spec);
        if (aliasEntry && aliasEntry.type === 'package') {
            const { from, to } = aliasEntry;
            spec = spec.replace(from, to);
        }
        // 2. parse spec
        const pkgInfo = this.parseSpec(spec);
        if (!pkgInfo)
            return spec;
        // 3. use resolve cache if present
        const { name: packageName, subpath } = pkgInfo;
        let importID = packageName + subpath.replace(/^\./, '');
        const cached = this.importMap.imports[importID];
        if (cached) {
            return this.pkgPrefix + cached.replace(/^\.\//, '');
        }
        // 4. resolve package, load deps
        const { manifest, cwd } = await this.findPkg(spec);
        const pkgID = `${packageName}@${manifest.version}`;
        const outdir = new URL(pkgID + '/', this.cacheDir);
        const external = new Set();
        // deps are external
        for (const dep of [
            ...Object.keys(manifest.dependencies || {}),
            ...Object.keys(manifest.peerDependencies || {}),
        ].filter((pkg) => !this.alwaysBundle.has(pkg))) {
            external.add(dep);
        }
        // so are already-built packages
        for (const key of Object.keys(this.importMap.imports)) {
            const pkgInfo = this.parseSpec(key);
            if (!pkgInfo || pkgInfo.name === packageName)
                continue; // don’t mark itself as external
            external.add(pkgInfo.name);
        }
        // 5. build core pkg
        const entryPoints = await this.findEntryPoints(manifest, { cwd });
        const esbuildEntryPoints = {};
        for (const [k, v] of Object.entries(entryPoints)) {
            esbuildEntryPoints[k === '.' ? 'index' : k.replace(/^\.\//, '').replace(/\.js$/, '')] = url_1.default.fileURLToPath(v);
        }
        await this.buildPackage({
            bundle: true,
            define: {
                'process.env.NODE_ENV': `"${this.mode}"`,
            },
            entryPoints: esbuildEntryPoints,
            external: [...external],
            format: this.target === 'node-cjs' ? 'cjs' : 'esm',
            minify: this.mode === 'production',
            outdir: url_1.default.fileURLToPath(outdir),
            platform: 'node',
            target: this.target.startsWith('node-') ? 'node16' : 'es2019',
            write: true,
        });
        // 6. add all entryPoints to cache
        const localImports = {};
        this.importMap.imports[packageName] = `./${pkgID}/index.js`;
        this.importMap.imports[pkgID] = `./${pkgID}/index.js`;
        this.importMap.imports[`${pkgID}/index.js`] = `./${pkgID}/index.js`;
        localImports[packageName] = `./index.js`;
        localImports[pkgID] = `./index.js`;
        localImports[`${pkgID}/index.js`] = `./index.js`;
        for (const k of Object.keys(entryPoints)) {
            const key = k === '.' ? 'index' : k.replace(/^\.\//, '');
            const value = path_1.default.extname(key) ? key : key + '.js';
            this.importMap.imports[`${packageName}/${key}`] = `./${pkgID}/${value}`;
            this.importMap.imports[`${pkgID}/${key}`] = `./${pkgID}/${value}`;
            localImports[`${packageName}/${key}`] = `./${value}`;
            localImports[`${pkgID}/${key}`] = `./${value}`;
            if (!path_1.default.extname(key)) {
                this.importMap.imports[`${packageName}/${key}.js`] = `./${pkgID}/${value}`;
                this.importMap.imports[`${pkgID}/${key}.js`] = `./${pkgID}/${value}`;
                localImports[`${packageName}/${key}.js`] = `./${value}`;
                localImports[`${pkgID}/${key}.js`] = `./${value}`;
            }
        }
        // 7. process dependencies
        await Promise.all([...external].map((pkg) => this.resolvePackageImport(pkg)));
        // 8. write import-map.json
        const importKeys = Object.keys(this.importMap.imports);
        importKeys.sort((a, b) => a.localeCompare(b, 'en-us', { numeric: true }));
        const sortedImports = {};
        for (const k of importKeys) {
            sortedImports[k] = this.importMap.imports[k];
        }
        await fs_1.default.writeFileSync(new URL('import-map.json', this.cacheDir), JSON.stringify({ ...this.importMap, imports: sortedImports }, undefined, 2), 'utf8');
        await fs_1.default.writeFileSync(new URL('import-map.json', outdir), JSON.stringify({ imports: localImports }, undefined, 2), 'utf8');
        // 9. finish
        const resolved = `${this.pkgPrefix}${pkgID}${subpath === '.' ? '/index.js' : subpath}`;
        return resolved;
    }
    async modifyBuildInstallOptions(opts) {
        return opts;
    }
    getCacheFolder() {
        return url_1.default.fileURLToPath(this.cacheDir);
    }
    async clearCache() {
        await del_1.default(url_1.default.fileURLToPath(this.cacheDir));
    }
    // internal
    /** Given an import specifier, return npm package info, or "undefined" if it’s non-npm */
    parseSpec(spec) {
        if (spec[0] === '.' || spec[0] === '/' || spec.includes('://'))
            return undefined; // relative spec, or remote spec
        let name = spec;
        // namespace
        let namespace;
        let namespaceMatch = name.match(/([a-zA-Z]+):(.*)/);
        if (namespaceMatch && namespaceMatch.length === 3) {
            namespace = namespaceMatch[1];
            name = namespaceMatch[2];
        }
        // scope
        let isScoped = false;
        if (name[0] === '@')
            isScoped = true;
        // subpath
        let parts = name.split('/');
        name = parts.shift();
        if (isScoped) {
            name = parts.shift();
        }
        let subpath = parts.join('/');
        if (subpath)
            subpath = '/' + subpath;
        if (!subpath)
            subpath = '.';
        return {
            name,
            namespace,
            isScoped,
            subpath: subpath,
        };
    }
    /** Build package (with some retry logic) */
    async buildPackage(options) {
        try {
            await esbuild.build({ ...options });
        }
        catch (errors) {
            // retry scenario 1: if an external package isn’t marked in dependencies, try a rebuild with that
            const [err] = errors.errors || [];
            if (err.text && err.text.includes('mark it as external to exclude it from the bundle')) {
                const packageMatch = err.text.match(/Could not resolve "([^"]+)"/);
                const missingPkg = this.parseSpec(packageMatch[1] || '');
                if (missingPkg && !(options.external || []).includes(missingPkg.name)) {
                    await this.resolvePackageImport(missingPkg.name); // build package externally
                    return this.buildPackage({
                        ...options,
                        external: [...(options.external || []), missingPkg.name],
                    });
                }
            }
            throw err;
        }
    }
    /** Get a package.json manifest and cwd for any spec */
    async findPkg(spec) {
        let searchLoc = require.resolve(spec);
        let foundPkgManifests = [];
        let pkgManifestLoc = path_1.default.join(searchLoc, 'package.json');
        while (searchLoc !== path_1.default.dirname(searchLoc)) {
            if (fs_1.default.existsSync(pkgManifestLoc))
                foundPkgManifests.push(pkgManifestLoc);
            searchLoc = path_1.default.dirname(searchLoc);
            pkgManifestLoc = path_1.default.join(searchLoc, 'package.json');
        }
        if (!foundPkgManifests.length) {
            throw new Error(`Could not locate package.json for "${spec}"`);
        }
        const foundPkgs = await Promise.all(foundPkgManifests.map(async (manifestLoc) => ({
            manifest: JSON.parse(await fs_1.default.promises.readFile(manifestLoc, 'utf8')),
            cwd: new URL(`file://${path_1.default.dirname(manifestLoc)}/`),
        })));
        for (const { manifest, cwd } of foundPkgs) {
            if (manifest.name && manifest.version) {
                return { manifest, cwd };
            }
        }
        throw new Error(`Could not locate root package.json for "${spec}"`);
    }
    /** Find any npm package’s entrypoints (similar to esinstall’s logic, but moved here to migrate away from esinstall) */
    async findEntryPoints(manifest, { cwd }) {
        const entryPoints = {};
        // filter out certain files
        function skipFile(filePath) {
            if (!ESBUILD_EXTENSIONS.has(path_1.default.extname(filePath)))
                return true;
            if (filePath.endsWith('.spec.js') || filePath.endsWith('.test.js'))
                return true;
            if (filePath.includes(`${path_1.default.sep}svelte-native${path_1.default.sep}`))
                return true; // TODO: temp fix for svelte-hmr
            return false;
        }
        // handle value from exports map
        async function parseSubpath(k, v) {
            const isGlob = v.includes('*');
            if (!isGlob) {
                return { [k]: new URL(v, cwd) };
            }
            const urls = {};
            const [kPrefix, kSuffix] = k.split('*');
            const [vPrefix, vSuffix] = v.split('*');
            const files = await tiny_glob_1.default(v, { cwd: url_1.default.fileURLToPath(cwd), filesOnly: true });
            for (const f of files) {
                if (skipFile(f))
                    continue; // if esbuild can’t process this, skip
                let expandedKey = f.replace(vPrefix, kPrefix); // "*" must have prefix
                if (kSuffix && vSuffix)
                    expandedKey = expandedKey.replace(vSuffix, kSuffix); // "*" may not have suffix; check first
                urls[expandedKey] = new URL(`./${f}`, cwd);
            }
            return urls;
        }
        // get correct target
        function findTarget(subpath, targets) {
            for (const target of targets) {
                if (subpath[target])
                    return target;
            }
        }
        // scenario 1: export map
        if (manifest.exports) {
            // glob searching takes time, so parse every key in parallel
            // (order isn’t preserved, but that’s alright)
            await Promise.all(Object.entries(manifest.exports).map(async ([k, v]) => {
                // string
                if (typeof v === 'string') {
                    for (const [sk, sv] of Object.entries(await parseSubpath(k, v))) {
                        entryPoints[sk] = sv;
                    }
                    return;
                }
                // object
                let match = v.default; // default
                if (this.target === 'browser')
                    match = findTarget(v, [...BROWSER_SUBPATHS]);
                if (this.target === 'node-esm')
                    match = findTarget(v, [...NODE_ESM_SUBPATHS]);
                if (this.target === 'node-cjs')
                    match = findTarget(v, [...NODE_CJS_SUBPATHS]);
                if (!match)
                    return; // if subpath/target mismatch, silently fail (should we throw instead?)
                for (const [sk, sv] of Object.entries(await parseSubpath(k, v[match]))) {
                    entryPoints[sk] = sv;
                }
            }));
        }
        // scenario 2: no export map: scan entire package
        else {
            // TODO: only scan main entry file for CJS
            const files = await tiny_glob_1.default('**/*', { cwd: url_1.default.fileURLToPath(cwd), filesOnly: true });
            for (const f of files) {
                if (skipFile(f))
                    continue; // if esbuild can’t process this, skip
                entryPoints[`./${f}`] = new URL(f, cwd);
            }
        }
        // lastly, add main entrypoint if missing
        if (!entryPoints['.']) {
            switch (this.target) {
                case 'node-cjs': {
                    for (const k of ['main']) {
                        if (typeof manifest[k] === 'string')
                            entryPoints['.'] = new URL(manifest[k], cwd);
                    }
                    break;
                }
                case 'browser':
                default: {
                    for (const k of [
                        'browser',
                        'browser:module',
                        'module',
                        'main:exnext',
                        'jsnext:main',
                        'main',
                    ]) {
                        if (typeof manifest[k] === 'string')
                            entryPoints['.'] = new URL(manifest[k], cwd);
                    }
                    // "browser" may also be an object
                    if (typeof manifest.browser === 'object') {
                        for (const k of ['./', './index', './index.js']) {
                            if (manifest.browser[k]) {
                                entryPoints['.'] = new URL(manifest.browser[k], cwd);
                                break;
                            }
                        }
                    }
                    break;
                }
            }
        }
        return !Object.keys(entryPoints).length ? { '.': new URL('./index.js', cwd) } : entryPoints;
    }
    /** Parse & resolve package imports */
    async transformImports({ filename, contents, }) {
        const ext = path_1.default.extname(filename.pathname);
        const isJS = ext === '.js' || ext === '.mjs' || ext === '.cjs';
        if (!isJS)
            return contents;
        let code = Buffer.isBuffer(contents) ? contents.toString('utf8') : contents;
        try {
            let imports = [...es_module_lexer_1.default.parse(code)[0]].filter(({ n }) => this.parseSpec(`${n}`));
            // step 1: find any missing/hidden deps, and resolve
            await Promise.all(imports.map(async ({ n }) => {
                const pkgInfo = this.parseSpec(n);
                if (pkgInfo && pkgInfo.name && !this.importMap.imports[n])
                    await this.resolvePackageImport(n);
            }));
            // step 2: transform the imports (which are already cached, so there’s no waiting here)
            while (imports.length) {
                const next = imports.find(({ n }) => this.importMap.imports[`${n}`]);
                if (!next)
                    break;
                const resolved = await this.resolvePackageImport(next.n);
                code = code.substring(0, next.s) + resolved + code.substring(next.e);
                imports = [...es_module_lexer_1.default.parse(code)[0]].filter(({ n }) => this.importMap.imports[`${n}`]); // because we modified the file, we must re-scan
            }
            return code;
        }
        catch (err) {
            try {
                const imports = cjs_module_lexer_1.default.parse(code);
                console.log({ cjs: imports }); // TODO: resolve this
                return code;
            }
            catch (err) {
                // give up
            }
        }
        return contents;
    }
}
exports.PackageSourceV4 = PackageSourceV4;
